"""
Overview:
    Meta information for imgutils package.
"""

#: Title of this project (should be `imgutils`).
__TITLE__ = 'imgutils'

#: Version of this project.
__VERSION__ = '0.9.1'

#: Short description of the project, will be included in ``setup.py``.
__DESCRIPTION__ = 'A convenient and user-friendly anime-style image data processing library that integrates ' \
                  'various advanced anime-style image processing models.'

#: Author of this project.
__AUTHOR__ = 'narugo1992, 7eu7d7'

#: Email of the authors'.
__AUTHOR_EMAIL__ = 'narugo1992@deepghs.org'
